#include "qarchive.h"
#include "bobbuilder.h"
#include "configutils.h"
#include "environment.h"
#include "fileunpack.h"
#include "finishpage.h"
#include "introscreen.h"
#include "keyinfo.h"
#include "license.h"
#include "licensepage.h"
#include "optionspage.h"
#include "qsainstaller.h"
#include "resource.h"
#include "shell.h"
#include "sidedecoration.h"
#include "uninstaller.h"

#include <qapplication.h>
#include <qcheckbox.h>
#include <qcombobox.h>
#include <qdir.h>
#include <qfiledialog.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qlineedit.h>
#include <qmessagebox.h>
#include <qpushbutton.h>
#include <qprocess.h>
#include <qprogressbar.h>
#include <qradiobutton.h>
#include <qtextedit.h>
#include <qtimer.h>
#include <qtoolbutton.h>
#include <qregexp.h>

static const char *introTextInstall =
    "<html><head><meta name=\"qrichtext\" content=\"1\" /></head>"
    "<body style=\"font-size:8pt;font-family:MS Shell Dlg\">"
    "<p>This wizard will install Qt Script for Applications (QSA) on your computer</p>"
    "<p>QSA requires Qt (3.1.2 or newer version) to be installed prior to installing QSA."
    "Make sure that the <code>QTDIR</code> environment variable points to the Qt"
    "installation directory and  that <code>QTDIR/bin</code> is in your PATH environment"
    "variable. For the installer to select the correct compiler, the <code>QMAKESPEC"
    "</code> environment variable should also This is already done if you installed Qt"
    "using the Qt installation wizard. </p>"
    "<p>Thank you for choosing QSA</p></body></html>";

static const char *introTextUninstall =
    "<html><head><meta name=\"qrichtext\" content=\"1\" /></head>"
    "<body style=\"font-size:8pt;font-family:MS Shell Dlg\">"
    "<p>This wizard will uninstall Qt Script for Applications (QSA) from your computer</p>"
    "<p>Press <code>Next</code> to proceed</body></html>";

static const char *pageTitlesInstall[] = {
    "Introduction",
    "License information",
    "License agreement",
    "Choose options",
    "Install files",
    "Build QSA",
    "Finish",
    0
};

static const char *pageTitlesUninstall[] = {
    "Introduction",
    "Uninstall",
    0
};

static QString createDecore( int name, const char *pageTitles[] )
{
    int count = -1;
    QString str;
    while( pageTitles[++count] ) {
        if( count == name ) {
            str += "<b>" + QString( pageTitles[count] ) + "</b><br>";
	}
        else {
            str += QString( pageTitles[count] ) + "<br>";
	}
    }
    return str;
}

#define addQSAPage( type, name, assign, widget ) \
{ \
    widget = new QWidget( this ); \
    QBoxLayout *layout = new QBoxLayout( widget, QBoxLayout::LeftToRight ); \
    SideDecoration *side = new SideDecoration( widget ); \
    side->text->setText( createDecore( name, pageTitles ) ); \
    layout->addWidget( side ); \
    assign = new type( widget ); \
    layout->addWidget( assign ); \
    addPage( widget, pageTitles[name] ); \
    setHelpEnabled( widget, FALSE ); \
}


QSAInstaller::QSAInstaller( QWidget *parent, const char *name, InstallerType type )
    : QWizard( parent, name ), unpackProgressLimit( 0 ), makeProcess( 0 ), installerType( type )
{
    shell = new WinShell();

    if( type == InstallType ) {
	pageTitles = pageTitlesInstall;

	addQSAPage( IntroScreen,  0, intro,       introPage );
	addQSAPage( LicensePage,  1, licenseInfo, licenseInfoPage );
	addQSAPage( License,      2, license,     licensePage );
	addQSAPage( OptionsPage,  3, options,     optionsPage );
	addQSAPage( FileUnpack,   4, unpack,      unpackPage );
	addQSAPage( BobBuilder,   5, build,       buildPage );
	addQSAPage( FinishPage,   6, finish,      finishPage );

	setNextEnabled( licensePage, FALSE );
	setNextEnabled( licenseInfoPage, FALSE );
	setNextEnabled( unpackPage, FALSE );
	setNextEnabled( buildPage, FALSE );
	setBackEnabled( buildPage, FALSE );

	connect( this, SIGNAL( selected( const QString & ) ),
		 this, SLOT( pageChangeInstaller( const QString & ) ) );
	connect( license->radioAgree, SIGNAL( toggled( bool ) ),
		 this, SLOT( toggleLicensePage( bool ) ) );
	connect( options->buttonDestPath, SIGNAL( clicked() ),
		 this, SLOT( findDirectory() ) );
	connect( options->buttonFolderPath, SIGNAL( clicked() ),
		 this, SLOT( findFolder() ) );
	connect( options->installPath, SIGNAL( textChanged( const QString & ) ),
		 this, SLOT( installPathChanged( const QString & ) ) );
	connect( licenseInfo->readLicenseButton, SIGNAL( clicked() ),
		 this, SLOT( readLicenseFile() ) );
	connect( licenseInfo->key, SIGNAL( textChanged( const QString & ) ),
		 this, SLOT( licenseKeyChanged( const QString & ) ) );

	fillLicenseInfo( licenseFile() );
	intro->text->setText( introTextInstall );

    } else if ( type == UninstallType ) {
	pageTitles = pageTitlesUninstall;
	addQSAPage( IntroScreen, 0, intro, introPage );
	addQSAPage( Uninstaller, 1, uninstaller, uninstallerPage );

	setBackEnabled( uninstallerPage, FALSE );

	connect( this, SIGNAL( selected( const QString & ) ),
		 this, SLOT( pageChangeUninstaller( const QString & ) ) );

	intro->text->setText( introTextUninstall );
    }
}


QSAInstaller::~QSAInstaller()
{
    if( makeProcess ) {
	if( makeProcess->isRunning() )
	    makeProcess->kill();
	delete makeProcess;
    }
}


void QSAInstaller::verifyEnvironment()
{
    QString qtDir = QEnvironment::getEnv( "QTDIR" );
    QString spec = QEnvironment::getEnv( "QMAKESPEC" );

    if( qtDir.isEmpty() || spec.isEmpty() ) {
	QStringList error;
	error << "<p>The following environment variables required by QSA"
	    " do not appear to be set:</p>";
	if( qtDir.isEmpty() ) error << "<code>QTDIR</code><br>";
	if( spec.isEmpty() ) error << "<code>QMAKESPEC</code><br>";
	error << "<p>Please verify that Qt 3.1.2 or higher version is properly installed</p>";
	QMessageBox::critical( this, QString::null, error.join( "" ) );
	setNextEnabled( introPage, FALSE );
    }
}


void QSAInstaller::pageChangeInstaller( const QString &pageName )
{
    if( pageName == pageTitles[0] ) {	      // Intro
	setCaption( "QSA Installation Wizard - Step 1 of 7" );
	QTimer::singleShot( 20, this, SLOT( verifyEnvironment() ) );

    } else if( pageName == pageTitles[1] ) {  // License info
	setCaption( "QSA Installation Wizard - Step 2 of 7" );

    } else if( pageName == pageTitles[2] ) {  // License agreement
	setCaption( "QSA Installation Wizard - Step 3 of 7" );
	readLicenseText();
	writeLicenseFile();

    } else if( pageName == pageTitles[3] ) {  // Build options
	setCaption( "QSA Installation Wizard - Step 4 of 7" );

    } else if( pageName == pageTitles[4] ) {  // Unpack files
	setCaption( "QSA Installation Wizard - Step 5 of 7" );
	QTimer::singleShot( 20, this, SLOT( unpackFiles() ) );

    } else if( pageName == pageTitles[5] ) {  // Configure & build QSA
	setCaption( "QSA Installation Wizard - Step 6 of 7" );
	QTimer::singleShot( 20, this, SLOT( buildQSA() ) );

    } else if( pageName == pageTitles[6] ) {  // Finish
	setCaption( "QSA Installation Wizard - Step 7 of 7" );
	readReadmeText();

    }

}


void QSAInstaller::pageChangeUninstaller( const QString &pageName )
{
    if( pageName == pageTitles[0] ) { // intro
	setCaption( "QSA Installation Wizard - Step 1 of 2" );

    } else if( pageName == pageTitles[1] ) { // Uninstall
	setCaption( "QSA Installation Wizard - Step 2 of 2" );
	QTimer::singleShot( 20, this, SLOT( uninstallQSA() ) );

    }
}


void QSAInstaller::installPathChanged( const QString &txt )
{
    QRegExp re( "^[a-zA-Z]:(\\\\[\\w-_]+)+$" );
    Q_ASSERT( re.isValid() );

    setNextEnabled( optionsPage, re.search( txt )==0 );
}


void QSAInstaller::toggleLicensePage( bool enable )
{
    setNextEnabled( licensePage, enable );
}


static QString get_value( const QString &src )
{
    int start = src.find( "=" ) + 1;
    if( src[start] == "\"" ) ++start;

    int stop = src.endsWith( "\"" ) ? src.length() - 1 : src.length();

    return src.mid( start, stop-start );
}


void QSAInstaller::readLicenseFile()
{
    QString fileName = QFileDialog::getOpenFileName();
    if( fileName.isEmpty() )
	return;
    QFile file( fileName );
    if( !file.open( IO_ReadOnly ) ) {
	QMessageBox::warning( this,
			      QString::null,
			      "Failed to open license file" );
	return;
    }
    fillLicenseInfo( file.readAll() );
}


void QSAInstaller::fillLicenseInfo( const QString &l )
{
    if( l.find( "LicenseKey" ) == -1 ) {
	QMessageBox::warning( this,
			      QString::null,
			      "License file does not contain license info!" );
	return;
    }

    QStringList lst = QStringList::split( QRegExp( "\r|\n" ), l );
    for( QStringList::ConstIterator it = lst.begin();
	 it != lst.end(); ++it ) {
	if ( (*it).startsWith( "CustomerID" ) )
	    licenseInfo->customerID->setText( get_value( *it ) );
	else if ( (*it).startsWith( "LicenseID" ) )
	    licenseInfo->licenseID->setText( get_value( *it ) );
	else if( (*it).startsWith( "Licensee" ) )
	    licenseInfo->licenseeName->setText( get_value( *it ) );
	else if( (*it).startsWith( "Products" ) )
	    licenseInfo->productsString->setCurrentText( get_value( *it ) );
	else if( (*it).startsWith( "ExpiryDate" ) )
	    licenseInfo->expiryDate->setText( get_value( *it ) );
	else if( (*it).startsWith( "LicenseKey" ) )
	    licenseInfo->key->setText( get_value( *it ) );
    }
}


void QSAInstaller::licenseKeyChanged( const QString &str )
{
    QString text = str.stripWhiteSpace().upper();
    QRegExp re( "[0-9A-Z]{4}-[0-9A-Z]{4}-[0-9A-Z]{4}" );

    QString edate;
    int prodIndex = 0;
    int qsaIndex = 1;
    bool nextEnable = FALSE;

    if( re.search( text ) == 0 ) {
	if( text.length() > 10 )
	    edate = decodedExpiryDate( text.mid(9) ).toString( ISODate );
	uint flags = convertor( text.mid( 1, 3 ) + text.mid( 5, 4 ) );
	prodIndex = flags & Feature_Enterprise ? 1 : 0;
	qsaIndex = flags & Feature_Extra2 ? 0 : 1;
	int tmp = Feature_Enterprise | Feature_Extra2;
	nextEnable = ( flags & tmp ) == tmp;
    }

    licenseInfo->expiryDate->setText( edate );
    licenseInfo->productsString->setCurrentItem( prodIndex );
    licenseInfo->qsa->setCurrentItem( qsaIndex );
    setNextEnabled( licenseInfoPage, nextEnable );
}

void QSAInstaller::writeLicenseFile()
{
    QString home = QDir::homeDirPath();
    if( home.endsWith( "/" ) || home.endsWith( "\\" ) )
	home = home.mid( 0, home.length() - 1 );
    QFile file( home + "\\.qt-license" );
    if( !file.open( IO_WriteOnly ) ) {
	QMessageBox::warning( this, QString::null,
			      "Failed to write license file: "
			      + home + "\\.qt-license" );
	return;
    }

    QTextStream s( &file );
    s << "# Toolkit license file\n"
      << "CustomerID=\"" << licenseInfo->customerID->text() << "\"\n"
      << "LicenseID=\"" << licenseInfo->licenseID->text() << "\"\n"
      << "Licensee=\"" << licenseInfo->licenseeName->text() << "\"\n"
      << "Products=\"" << licenseInfo->productsString->currentText() << "\"\n"
      << "ExpiryDate=" << licenseInfo->expiryDate ->text() << "\n"
      << "LicenseKey=" << licenseInfo->key->text() << "\n";
}


void QSAInstaller::readLicenseText()
{
    if( !license->licenseOutput->text().isEmpty() )
	return;

    bool usLicense = isUsLicense();
    ResourceLoader loader( usLicense ? "LICENSE-US" : "LICENSE" );
    QString licenseText = loader.data();

    if( licenseText.isEmpty() ) {
	QMessageBox::critical( this,
			       "Installer failed",
			       "Package corrupt, could not locate LICENSE file" );
	qApp->quit();
    }

    license->licenseOutput->setText( licenseText );
}


void QSAInstaller::readReadmeText()
{
    if( !finish->readme->text().isEmpty() )
	return;

    QFile file( "README" );
    QString content;
    if( !file.open( IO_ReadOnly ) )
	content = "README file not found in package,\nplease report this to qsa@trolltech.com\n";
    else
	content = QString( file.readAll() );
    finish->readme->setText( content );
    setFinishEnabled( finishPage, TRUE );
}


void QSAInstaller::findDirectory()
{
    QString orig = QDir::convertSeparators( options->installPath->text() );
    QString dir = QFileDialog::getExistingDirectory( orig, this );
    if( !dir.isEmpty() )
	options->installPath->setText( dir );
}


void QSAInstaller::findFolder()
{
    int select = options->folderGroup->currentItem();
    QString folder = shell->selectFolder( options->programFolder->text(), select == 0 );
    options->programFolder->setText( folder );
}


void QSAInstaller::unpackFiles()
{
    if( unpackProgressLimit )
	return;

    ResourceLoader loader( "QT_ARQ", 500 );
    bool success = FALSE;
    if ( loader.isValid() ) {
	QArchive ar;
	ar.setVerbosity( QArchive::Progress );
	connect( &ar, SIGNAL( operationFeedback( int ) ),
		 unpack->progress, SLOT( setProgress( int ) ) );
	connect( &ar, SIGNAL( operationFeedback( int ) ),
		 this, SLOT( unpackProgress( int ) ) );
	unpack->progress->setTotalSteps( loader.data().count() );
	unpackProgressLimit = loader.data().count();
	QDataStream ds( loader.data(), IO_ReadOnly );
	success = ar.readArchive( &ds, options->installPath->text(), licenseInfo->key->text() );
    }

    if( !success ) {
	QMessageBox::critical( this, QString::null, "Unable to extract files" );
	qApp->quit();
    }

    unpack->label->setText( "Unpacking files done..." );

}


void QSAInstaller::unpackProgress( int progress )
{
    if( progress == unpackProgressLimit ) {
	copy( isUsLicense()
	      ? options->installPath->text() + "\\.LICENSE-US"
	      : options->installPath->text() + "\\.LICENSE",
	      options->installPath->text() + "\\LICENSE" );
	setNextEnabled( unpackPage, TRUE );
	next();
    }
}


#define exit_on_message( parent ) \
    if( messages().count() ) { \
	QMessageBox::critical( parent, QString::null, messages().join( "\n" ) ); \
	qApp->quit(); \
	return; \
    }


void QSAInstaller::buildQSA()
{
    static bool buildTried = FALSE;
    if( buildTried ) return;

    buildTried = TRUE;
    autoShutdown = FALSE;

    if( !QDir::setCurrent( options->installPath->text() ) ) {
	QMessageBox::critical( this, QString::null,
			       "Failed access the install directory!" );
	qApp->quit();
	return;
    }
    qtDir = new QString( QEnvironment::getEnv( "QTDIR" ) + QString( "\\" ) );

    copy( (*qtDir) + ".qmake.cache", ".qmake.cache" );

    if( !writeQSConfig( options->checkBuildIde->isChecked() ) ) {
	QMessageBox::critical( this, QString::null,
			      "Failed to write the QSA configuration file: <pre>"
			      + messages().join( "\n" ) + "</pre>" );
	qApp->quit();
	return;
    }

    copyQSAHeaders();
    exit_on_message( this );

    installDocs();
    if( messages().count() ) {
	QMessageBox::warning( this, QString::null, "Failed to install documentation" );
    }

    QStringList configs;
    if( !options->checkBuildIde->isChecked() )
	configs << "noide";

    runQMake( configs, QString::null );
    exit_on_message( this );

    QTimer::singleShot( 100, this, SLOT( runMake() ) );
}


void QSAInstaller::runMake()
{
    if( processes ) {
	QTimer::singleShot( 100, this, SLOT( runMake() ) );
	return;
    }

    QString qmakeSpec = QEnvironment::getEnv( "QMAKESPEC" );
    QString makeCommand = qmakeSpec.startsWith( "win32-msvc" ) ? "nmake" : "make";

    makeProcess = new QProcess( makeCommand, this );
    connect( makeProcess, SIGNAL( readyReadStdout() ), this, SLOT( makeRead() ) );
    connect( makeProcess, SIGNAL( readyReadStderr() ), this, SLOT( makeRead() ) );
    connect( makeProcess, SIGNAL( processExited() ), this, SLOT( makeDone() ) );

    if( !makeProcess->start() ) {
	QMessageBox::critical( this, QString::null,
			       "Failed to start build process" );
    }
}


void QSAInstaller::makeDone()
{
    if( makeProcess->exitStatus() == 0 ) {

	QString sout( makeProcess->readStdout() );
	if( !sout.isEmpty() )
	    build->buildOutput->append( sout );
	QString serr( makeProcess->readStderr() );
	if( !serr.isEmpty() )
	    build->buildOutput->append( serr );

	writeFolders();
	registerUninstall();
	setNextEnabled( buildPage, TRUE );
	build->buildOutput->append( "\nBuilding QSA complete..." );
    } else {
	QMessageBox::critical( this, QString::null, "Compilation failed..." );
    }
}

void QSAInstaller::makeRead()
{
    if( makeProcess->canReadLineStdout() )
	build->buildOutput->append( makeProcess->readLineStdout() );
    if( makeProcess->canReadLineStderr() )
	build->buildOutput->append( makeProcess->readLineStderr() );
}


void QSAInstaller::writeFolders()
{
    QString folder = options->programFolder->text();

    if( folder.isEmpty() )
	return;

    bool common = options->folderGroup->currentItem() == 0;
    folderPath = shell->createFolder( folder, common );

    QString examples = shell->createFolder( folder + "\\Examples", common );
    QString examplePath = options->installPath->text() + "\\examples\\";
    shell->createShortcut( examples, common, "Console", examplePath + "console\\console.exe" );
    shell->createShortcut( examples, common, "Filter", examplePath + "filter\\filter.exe" );
    shell->createShortcut( examples, common, "Game", examplePath + "game\\game.exe" );
    shell->createShortcut( examples, common, "ScribbleScripter", examplePath + "scribblescripter\\scribblescripter.exe" );
    shell->createShortcut( examples, common, "Scriptbutton", examplePath + "scriptbutton\\scriptbutton.exe" );
    shell->createShortcut( examples, common, "SpreadSheet", examplePath + "spreadsheet\\spreadsheet.exe" );
    shell->createShortcut( examples, common, "TextEdit", examplePath + "textedit\\textedit.exe" );

    shell->createShortcut( folderPath, common, "License agreement for QSA", "notepad.exe", "Review the license agreement",
			  "\"" + options->installPath->text() + "\\LICENSE\"" );
    shell->createShortcut( folderPath, common, "Readme for QSA", "notepad.exe", "Important information",
			  "\"" + options->installPath->text() + "\\README\"" );
    shell->createShortcut( folderPath, common, "On-line documentation", (*qtDir) + "bin\\assistant.exe",
			   "Browse the On-line documentation", "", *qtDir );
    shell->createInternetShortcut( folderPath, common, "Trolltech.com", "http://www.trolltech.com/" );
}

void QSAInstaller::registerUninstall()
{
    QString qsaDir = options->installPath->text();

    copy( qApp->argv()[0], shell->windowsFolderName + "\\qsauninstall.exe" );
    QString cmd = shell->windowsFolderName + "\\qsauninstall.exe -uninstall \"" + qsaDir + "\" \"" + folderPath + "\"";
    QEnvironment::recordUninstall( "QSA 1.0.0", cmd );
}




void QSAInstaller::uninstallQSA()
{
    QString qtDir = QEnvironment::getEnv( "QTDIR" );
    QString qmakeSpec = QEnvironment::getEnv( "QMAKESPEC" );

    QStringList files;
    files << qtDir + "\\include\\qsaglobal.h"
	  << qtDir + "\\include\\qsobjectfactory.h"
	  << qtDir + "\\include\\qswrapperfactory.h"
	  << qtDir + "\\include\\qseditor.h"
	  << qtDir + "\\include\\qsproject.h"
	  << qtDir + "\\include\\qsinterpreter.h"
	  << qtDir + "\\include\\qsargument.h"
	  << qtDir + "\\include\\qsinputdialogfactory.h"
	  << qtDir + "\\include\\qsscript.h"
	  << qtDir + "\\include\\qsconfig.h"
	  << qtDir + "\\include\\qsworkbench.h"
	  << qtDir + "\\bin\\qsa100.dll"
	  << qtDir + "\\lib\\qsa100.lib"
	  << qtDir + "\\lib\\qsa100.dll"
	  << qtDir + "\\plugins\\designer\\qseditorplugin.dll"
	  << qtDir + "\\plugins\\designer\\qseditorplugin.lib";
    for( QStringList::ConstIterator it = files.begin(); it != files.end(); ++it ) {
	QFile file( *it );
	if( file.exists() && file.remove() )
	    uninstaller->output->append( "Deleted file: " + file.name() );
	else
	    uninstaller->output->append( "<font color=red>Failed to delete file: "
					 + file.name() + "</font>" );
    }


    QString qsaDirStr = qApp->argv()[2];
    if( !qsaDirStr.isEmpty() ) {
	QDir qsaDir = QDir( qApp->argv()[2] );
	rmDirRecursive( qsaDir );
	if( qsaDir.exists() )
	    uninstaller->output->append( "<font color=red>Failed to delete directory: "
				     + qsaDir.absPath() + "</font>" );
	else
	    uninstaller->output->append( "Deleted directory: " + qsaDir.absPath() );
    }

    QString folderStr = qApp->argv()[3];
    if( !folderStr.isEmpty() ) {
	QDir folder = QDir( qApp->argv()[3] );
	rmDirRecursive( folder );
	if( folder.exists() )
	    uninstaller->output->append( "<font color=red>Failed to delete directory: "
					 + folder.absPath() + "</font>" );
	else
	    uninstaller->output->append( "Deleted directory: " + folder.absPath() );
    }

    QEnvironment::removeUninstall( QString( "QSA 1.0.0" ) );
    uninstaller->output->append( "<br>QSA has now been removed from your system..." );

    setFinishEnabled( uninstallerPage, TRUE );
}
